from .ply import yacc
from .lex import *
from .ast import *
from .builtins import *
from .exception import *


# Precedence rules
precedence = (
    ('left', 'OR', 'AND'),
    ('left', 'LESS', 'GREATER', 'LESSEQUAL', 'GREATEREQUAL', 'EQEQUAL', 'NOTEQUAL'),
    ('left', 'PLUS', 'MINUS'),
    ('left', 'STAR', 'SLASH', 'DOUBLESLASH', 'PERCENT'),
    ('left', 'DOUBLESTAR'),
    ('right', 'NOT', 'UADD', 'USUB')
)

# Rule for a program
def p_program(p):
    '''program : statements
               | empty'''
    body = []
    if (p[1]): body = p[1]
    p[0] = Program(body)

# Rule for statements
def p_statements(p):
    '''statements : statements statement
                  | statement'''
    if (len(p) == 2 and p[1]):
        statement = p[1]
        p[0] = [statement]
    elif (len(p) == 3):
        statements = p[1]
        if (p[2]): statements.append(p[2])
        p[0] = statements

# Statement rule for function definition
def p_statement_functiondef(p):
    '''statement : FUNCTIONDEF NAME LPAR parameters RPAR BEGIN statements END
                 | FUNCTIONDEF NAME LPAR RPAR BEGIN statements END'''
    if (len(p) == 8):
        name = Name(p[2])
        body = p[6]
        p[0] = FunctionDef(name, [], body)
    elif (len(p) == 9):
        name = Name(p[2])
        parameters = p[4]
        body = p[7]
        p[0] = FunctionDef(name, parameters, body)

# Rule for return
def p_statement_return(p):
    '''statement : RETURN expression'''
    p[0] = Return(p[2])

# Statement rule for assignment
def p_statement_assignment(p):
    '''statement : NAME EQUAL expression'''
    name = Name(p[1])
    p[0] = Assignment(name, p[3])

# Statement rule for if-then
def p_statement_ifthen(p):
    '''statement : IF expression THEN statements statement_elifs END'''
    expression = p[2]
    statements = p[4]
    statement_elifs = p[5]
    p[0] = If(expression, statements, statement_elifs)

# Statement rule for if-then-else
def p_statement_ifthenelse(p):
    '''statement : IF expression THEN statements statement_elifs ELSE statements END'''
    expression = p[2]
    statements_if = p[4]
    statements_else = p[7]
    statement_elifs = p[5]
    p[0] = If(expression, statements_if, statement_elifs, statements_else)

# Statement rule for elifs
def p_statement_elifs(p):
    '''statement_elifs : statement_elifs statement_elif
                       | statement_elif
                       | empty'''
    if (len(p) == 2 and p[1]):
        statement_elif = p[1]
        p[0] = [statement_elif]
    elif (len(p) == 3):
        statement_elifs = p[1]
        statement_elif = p[2]
        if (p[2]): statement_elifs.append(statement_elif)
        p[0] = statement_elifs
    else:
        p[0] = []

# Statement rule for elif
def p_statement_elif(p):
    '''statement_elif : ELIF expression THEN statements'''
    expression = p[2]
    statements = p[4]
    p[0] = Elif(expression, statements)

# Statement rule for iterate
def p_statement_iterate(p):
    '''statement : FOR NAME ITERATE NAME BEGIN statements END'''
    variable_identifier = p[2]
    iterable = Name(p[4])
    statements = p[6]
    p[0] = Iterate(variable_identifier, iterable, statements)

# Statement rule for loop
def p_statement_loop(p):
    '''statement : LOOP expression BEGIN statements END'''
    expression = p[2]
    statements = p[4]
    p[0] = Loop(expression, statements)

# Statement rule for continue
def p_statement_continue(p):
    '''statement : CONTINUE'''
    p[0] = Continue()

# Statement rule for break
def p_statement_break(p):
    '''statement : BREAK'''
    p[0] = Break()

# Statement rule for expression
def p_statement_expression(p):
    '''statement : expression'''
    p[0] = p[1]

# Rule for expressions
def p_expressions(p):
    '''expressions : expression comma_expressions COMMA
                   | expression comma_expressions
                   | expression COMMA
                   | expression'''
    if ((len(p) == 3 and p[2] == ',') or len(p) == 2):
        expression = p[1]
        p[0] = [expression]
    elif (len(p) == 4 or len(p) == 3):
        expressions = p[2]
        if (p[1]): expressions.insert(0, p[1])
        p[0] = expressions

# Rule for expressions: (',' expression)+
def p_comma_expressions(p):
    '''comma_expressions : comma_expressions COMMA expression
                         | COMMA expression'''
    if (len(p) == 3):
        expression = p[2]
        p[0] = [expression]
    elif (len(p) == 4):
        expressions = p[1]
        if (p[3]): expressions.append(p[3])
        p[0] = expressions

# Rule for parameters
def p_parameters(p):
    '''parameters : expressions'''
    for node in p[1]:
        if (type(node) != Name):
            raise SyntaxException(
                f'invalid parameter type: {type(node).__name__}'
            )
    p[0] = p[1]

# Expression rule for group surrounded with parentheses
def p_expression_group(p):
    '''expression : LPAR expression RPAR'''
    p[0] = p[2]

# Expression rule for unary operator
def p_expression_unary_operators(p):
    '''expression : NOT expression %prec NOT
                  | PLUS expression %prec UADD
                  | MINUS expression %prec USUB'''
    value = p[2]
    operator = EXACT_TOKEN_TYPES[p[1]]
    if (operator == 'NOT'):
        p[0] = Not(value)
    elif (operator == 'PLUS'):
        p[0] = UAdd(value)
    elif (operator == 'MINUS'):
        p[0] = USubtract(value)

# Expression rule for binary operator
def p_expression_binary_operators(p):
    '''expression : expression PLUS expression
                  | expression MINUS expression
                  | expression STAR expression
                  | expression SLASH expression
                  | expression PERCENT expression
                  | expression DOUBLESTAR expression
                  | expression DOUBLESLASH expression'''
    left = p[1]
    right = p[3]
    operator = EXACT_TOKEN_TYPES[p[2]]
    if (operator == 'PLUS'):
        p[0] = Add(left, right)
    elif (operator == 'MINUS'):
        p[0] = Subtract(left, right)
    elif (operator == 'STAR'):
        p[0] = Multiply(left, right)
    elif (operator == 'SLASH'):
        p[0] = Divide(left, right)
    elif (operator == 'PERCENT'):
        p[0] = Modulo(left, right)
    elif (operator == 'DOUBLESTAR'):
        p[0] = Power(left, right)
    elif (operator == 'DOUBLESLASH'):
        p[0] = FloorDivide(left, right)

# Expression rule for boolean operator
def p_expression_boolean_operators(p):
    '''expression : expression AND expression
                  | expression OR expression'''
    left = p[1]
    right = p[3]
    operator = EXACT_TOKEN_TYPES[p[2]]
    if (operator == 'AND'):
        p[0] = And(left, right)
    elif (operator == 'OR'):
        p[0] = Or(left, right)

# Expression rule for compare operator
def p_expression_compare_operators(p):
    '''expression : expression EQEQUAL expression
                  | expression NOTEQUAL expression
                  | expression LESS expression
                  | expression LESSEQUAL expression
                  | expression GREATER expression
                  | expression GREATEREQUAL expression'''
    left = p[1]
    right = p[3]
    operator = EXACT_TOKEN_TYPES[p[2]]
    if (operator == 'EQEQUAL'):
        p[0] = Equal(left, right)
    elif (operator == 'NOTEQUAL'):
        p[0] = NotEqual(left, right)
    elif (operator == 'LESS'):
        p[0] = Less(left, right)
    elif (operator == 'LESSEQUAL'):
        p[0] = LessEqual(left, right)
    elif (operator == 'GREATER'):
        p[0] = Greater(left, right)
    elif (operator == 'GREATEREQUAL'):
        p[0] = GreaterEqual(left, right)

# Expression rule for boolean constant
def p_expression_boolean(p):
    '''expression : TRUE
                  | FALSE'''
    constant = EXACT_TOKEN_TYPES[p[1]]
    if (constant == 'TRUE'):
        p[0] = Boolean(1)
    elif (constant == 'FALSE'):
        p[0] = Boolean(0)

# Expression rule for integer number
def p_expression_integer(p):
    '''expression : INTEGER'''
    p[0] = Integer(p[1])

# Expression rule for float number
def p_expression_float(p):
    '''expression : FLOAT'''
    p[0] = Float(p[1])

# Expression rule for complex number
def p_expression_complex(p):
    '''expression : COMPLEX'''
    p[0] = Complex(p[1])

# Expression rule for string
def p_expression_string(p):
    '''expression : STRING'''
    p[0] = String(p[1])

# Expression rule for tuple
def p_expression_tuple(p):
    '''expression : LPAR expressions RPAR
                  | LPAR RPAR'''
    if (len(p) == 3):
        p[0] = Tuple([])
    elif (len(p) == 4):
        expressions = p[2]
        p[0] = Tuple(expressions)

# Expression rule for list
def p_expression_list(p):
    '''expression : LSQB expressions RSQB
                  | LSQB RSQB'''
    if (len(p) == 3):
        p[0] = List([])
    elif (len(p) == 4):
        expressions = p[2]
        p[0] = List(expressions)

# Expression rule for dict
def p_expression_dict(p):
    '''expression : LBRACE key_value_pairs RBRACE
                  | LBRACE RBRACE'''
    if (len(p) == 3):
        p[0] = Dict([])
    elif (len(p) == 4):
        key_value_pairs = p[2]
        p[0] = Dict(key_value_pairs)

# Rule for key-value pairs
def p_key_value_pairs(p):
    '''key_value_pairs : key_value_pairs COMMA key_value_pair
                       | key_value_pair'''
    if (len(p) == 2 and p[1]):
        key_value_pair = p[1]
        p[0] = [key_value_pair]
    elif (len(p) == 4):
        key_value_pairs = p[1]
        if (p[3]): key_value_pairs.append(p[3])
        p[0] = key_value_pairs

# Rule for key-value pair
def p_key_value_pair(p):
    '''key_value_pair : expression COLON expression'''
    expression_key = p[1]
    expression_value = p[3]
    p[0] = KeyValuePair(expression_key, expression_value)

# Expression rule for iterable index
def p_expression_iterable_index(p):
    '''expression : NAME INDEX expression'''
    name = Name(p[1])
    expression = p[3]
    p[0] = IterableIndex(name, expression)

# Expression rule for null
def p_expression_null(p):
    '''expression : NULL'''
    p[0] = Null()

# Expression rule for variable name
def p_expression_name(p):
    '''expression : NAME'''
    p[0] = Name(p[1])

# Expression rule for function call
def p_expression_functioncall(p):
    '''expression : NAME LPAR expressions RPAR
                  | NAME LPAR RPAR'''
    if (len(p) == 4):
        name = Name(p[1])
        p[0] = FunctionCall(name, [])
    elif (len(p) == 5):
        name = Name(p[1])
        expressions = p[3]
        p[0] = FunctionCall(name, expressions)

# Rule for nothing
def p_empty(p):
    '''empty : '''

# Error rule for syntax error
def p_error(p):
    raise SyntaxException(f'invalid token {p.value}, line {p.lineno}')

# Build the parser
parser = yacc.yacc()



import sys
import inspect
from .lex import *
from .parse import *
from .ast import *
from .symtable import *
from .builtins import *


def exec(data, debug=False):
    if (debug):
        print('[DEBUG] input data as raw string:')
        print(repr(data))
        print('[DEBUG] input data as string:')
        print(data)

    # Lex
    lexer.input(data)

    if (debug):
        print('[DEBUG] tokens:')
        while (True):
            tok = lexer.token()
            if (not tok): break
            print(tok)
    
    if (debug):
        print('[DEBUG] program output:')
    
    # Initialise symbol table
    symtable = SymbolTable()

    # Add builtin functions to the symbol table
    builtins = {
        name: obj 
        for name, obj 
        in inspect.getmembers(sys.modules['src.sheng.builtins'])
    }
    for identifier, builtin_name in BUILTIN_NAMES.items():
        name = Name(identifier)
        function_pointer = builtins[builtin_name]
        parameters = [
            Name(arg) 
            for arg in inspect.getfullargspec(function_pointer)[0]
        ]
        functiondef = FunctionDef(name, parameters, function_pointer)
        symtable.insert(identifier, functiondef)
    
    # Parse
    ast = parser.parse(data)
    ast.evaluate(symtable)



BUILTIN_NAMES = {
    '打印': 'builtin_print',
    '绝对值': 'builtin_abs',
    '最大值': 'builtin_max',
    '最小值': 'builtin_min',
    '类型': 'builtin_type'
}


# -----------------------------------------------------------------------------
# Build-in types
# -----------------------------------------------------------------------------

class BaseType(object):

    def __init__(self, type_name):
        self.type_name = type_name

    def get_type(self):
        return f'<类型 \'{self.type_name}\'>'


class NullType(BaseType):

    def __init__(self, value):
        super().__init__('空类型')
        self.value = value

    def __repr__(self):
        return '空'

    def __str__(self):
        return '空'

    def __bool__(self):
        return bool(None)

    def __eq__(self, other):
        return other == None

    def __ne__(self, other):
        return other != None


class BooleanType(BaseType, int):

    def __init__(self, value):
        super().__init__('布林')
        self.value = value

    def __repr__(self):
        return '对' if self.value else '错'

    def __str__(self):
        return '对' if self.value else '错'

    def __bool__(self):
        return self.value


class IntegerType(BaseType, int):

    def __init__(self, value):
        super().__init__('整数')
        self.value = value


class FloatType(BaseType, float):

    def __init__(self, value):
        super().__init__('浮点数')
        self.value = value


class ComplexType(BaseType, complex):

    def __init__(self, value):
        super().__init__('复数')
        self.value = value


class StringType(BaseType, str):

    def __init__(self, value):
        super().__init__('字符串')
        self.value = value


class TupleType(BaseType, tuple):

    def __init__(self, value):
        super().__init__('元组')
        self.value = value
    
    def __new__(self, value):
        return tuple.__new__(self, value)


class ListType(BaseType, list):

    def __init__(self, value):
        super().__init__('列表')
        self.extend(value)


class DictType(BaseType, dict):

    def __init__(self, value):
        super().__init__('字典')
        self.update(value)


# -----------------------------------------------------------------------------
# Built-in functions
# -----------------------------------------------------------------------------

def builtin_print(value):
    print(value)
    return None

def builtin_abs(value):
    return abs(value)

def builtin_max(iterable):
    return max(iterable)

def builtin_min(iterable):
    return min(iterable)

def builtin_type(value):
    return value.get_type()



from abc import *
from .symtable import *
from .exception import *
from .builtins import *


BASE_TYPE_CLASSES = {
    'NoneType': NullType,
    'bool': BooleanType,
    'int': IntegerType,
    'float': FloatType,
    'complex': ComplexType,
    'str': StringType,
    'tuple': TupleType,
    'list': ListType,
    'dict': DictType
}


class ASTNode(ABC):

    def evaluate(self, symtable):
        raise NotImplementedError()


class ASTUtil(object):

    def evaluate_base_type(value):
        type_name = type(value).__name__
        base_types = [e.__name__ for e in BASE_TYPE_CLASSES.values()]
        if (type_name in base_types):
            return value
        elif (type_name in BASE_TYPE_CLASSES):
            base_type = BASE_TYPE_CLASSES[type_name]
            return base_type(value)
        else:
            raise RuntimeException(
                f'bad operand type: {type_name}'
            )

    def evaluate_statements(symtable, sequence):
        for node in sequence:
            value = node.evaluate(symtable)
            if (value):
                return value
        return None


class Program(ASTNode):

    def __init__(self, statements):
        self.statements = statements

    def evaluate(self, symtable):
        return ASTUtil.evaluate_statements(
            symtable, self.statements
        )


class Statement(ASTNode): pass


class FunctionDef(Statement, BaseType):

    def __init__(self, name, parameters, body):
        super().__init__('函数')
        self.name = name
        self.parameters = parameters
        self.body = body

    def evaluate(self, symtable):
        identifier = self.name.get_identifier()
        symbol = symtable.lookup(identifier)
        if (symbol):
            raise RuntimeException(f'{identifier} has been defined')
        symtable.insert(identifier, self)
        return None

    def get_parameters(self):
        return self.parameters

    def get_body(self):
        return self.body


class Return(Statement):

    def __init__(self, expression):
        self.expression = expression

    def evaluate(self, symtable):
        value = self.expression.evaluate(symtable)
        ReturnValue = type('ReturnValue', (type(value), ), {})
        return ReturnValue(value)


class Assignment(Statement):

    def __init__(self, name, node):
        self.name = name
        self.node = node

    def evaluate(self, symtable):
        identifier = self.name.get_identifier()
        node = Constant(self.node.evaluate(symtable))
        symtable.insert(identifier, node)
        return None


class If(Statement):

    def __init__(self, expression, statements, statement_elifs, statements_else=None):
        self.expression = expression
        self.statements = statements
        self.statement_elifs = statement_elifs
        self.statements_else = statements_else

    def evaluate(self, symtable):
        if (self.expression.evaluate(symtable)):
            return ASTUtil.evaluate_statements(
                symtable, self.statements
            )
        else:
            for statement_elif in self.statement_elifs:
                if (statement_elif.is_true(symtable)):
                    return statement_elif.evaluate(symtable)
            if (self.statements_else):
                return ASTUtil.evaluate_statements(
                    symtable, self.statements_else
                )
            return None


class Elif(Statement):

    def __init__(self, expression, statements):
        self.expression = expression
        self.statements = statements

    def is_true(self, symtable):
        return self.expression.evaluate(symtable)

    def evaluate(self, symtable):
        for node in self.statements:
            value = node.evaluate(symtable)
            if (value):
                return value
        return None


class Iterate(Statement):

    def __init__(self, variable_identifier, iterable, statements):
        self.variable_identifier = variable_identifier
        self.iterable = iterable
        self.statements = statements

    def evaluate(self, symtable):
        for e in self.iterable.evaluate(symtable):
            node = Constant(e)
            symtable.insert(self.variable_identifier, node)
            value = ASTUtil.evaluate_statements(
                symtable, self.statements
            )
            if (type(value).__name__ == 'ReturnValue'):
                return value
            if (type(value).__name__ == 'ContinueValue'):
                continue
            if (type(value).__name__ == 'BreakValue'):
                break
        return None


class Loop(Statement):

    def __init__(self, expression, statements):
        self.expression = expression
        self.statements = statements

    def evaluate(self, symtable):
        while (self.expression.evaluate(symtable)):
            value = ASTUtil.evaluate_statements(
                symtable, self.statements
            )
            if (type(value).__name__ == 'ReturnValue'):
                return value
            if (type(value).__name__ == 'ContinueValue'):
                continue
            if (type(value).__name__ == 'BreakValue'):
                break
        return None


class Continue(Statement):

    def evaluate(self, symtable):
        ContinueValue = type('ContinueValue', (object, ), {})
        return ContinueValue()


class Break(Statement):

    def evaluate(self, symtable):
        BreakValue = type('BreakValue', (object, ), {})
        return BreakValue()


class Expression(ASTNode): pass


class Constant(Expression):

    def __init__(self, value):
        self.value = value

    def evaluate(self, symtable):
        return self.value


class Boolean(Constant):

    def evaluate(self, symtable):
        return ASTUtil.evaluate_base_type(bool(self.value))


class Integer(Constant):

    def evaluate(self, symtable):
        return ASTUtil.evaluate_base_type(int(self.value))


class Float(Constant):

    def evaluate(self, symtable):
        return ASTUtil.evaluate_base_type(float(self.value))


class Complex(Constant):

    def evaluate(self, symtable):
        return ASTUtil.evaluate_base_type(complex(self.value))


class String(Constant):

    def evaluate(self, symtable):
        return ASTUtil.evaluate_base_type(str(self.value[1:-1]))


class Null(Expression):

    def evaluate(self, symtable):
        return ASTUtil.evaluate_base_type(None)


class UnaryOp(Expression):

    def __init__(self, node):
        self.node = node


class Not(UnaryOp):

    def evaluate(self, symtable):
        value = self.node.evaluate(symtable)
        return ASTUtil.evaluate_base_type(not value)


class UAdd(UnaryOp):

    def evaluate(self, symtable):
        value = self.node.evaluate(symtable)
        return ASTUtil.evaluate_base_type(+value)


class USubtract(UnaryOp):

    def evaluate(self, symtable):
        value = self.node.evaluate(symtable)
        return ASTUtil.evaluate_base_type(-value)


class BinaryOp(Expression):

    def __init__(self, left, right):
        self.left = left
        self.right = right


class Add(BinaryOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return ASTUtil.evaluate_base_type(lvalue + rvalue)


class Subtract(BinaryOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return ASTUtil.evaluate_base_type(lvalue - rvalue)


class Multiply(BinaryOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return ASTUtil.evaluate_base_type(lvalue * rvalue)


class Divide(BinaryOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return ASTUtil.evaluate_base_type(lvalue / rvalue)


class Modulo(BinaryOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return ASTUtil.evaluate_base_type(lvalue % rvalue)


class Power(BinaryOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return ASTUtil.evaluate_base_type(lvalue ** rvalue)


class FloorDivide(BinaryOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return ASTUtil.evaluate_base_type(lvalue // rvalue)


class BoolOp(BinaryOp): pass


class And(BoolOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return ASTUtil.evaluate_base_type(lvalue and rvalue)


class Or(BoolOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return ASTUtil.evaluate_base_type(lvalue or rvalue)


class CompareOp(BinaryOp): pass


class Equal(CompareOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return ASTUtil.evaluate_base_type(lvalue == rvalue)


class NotEqual(CompareOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return ASTUtil.evaluate_base_type(lvalue != rvalue)


class Less(CompareOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return ASTUtil.evaluate_base_type(lvalue < rvalue)


class LessEqual(CompareOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return ASTUtil.evaluate_base_type(lvalue <= rvalue)


class Greater(CompareOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return ASTUtil.evaluate_base_type(lvalue > rvalue)


class GreaterEqual(CompareOp):

    def evaluate(self, symtable):
        lvalue = self.left.evaluate(symtable)
        rvalue = self.right.evaluate(symtable)
        return ASTUtil.evaluate_base_type(lvalue >= rvalue)


class Name(Expression):

    def __init__(self, value):
        self.value = value

    def get_identifier(self):
        return self.value
    
    def evaluate(self, symtable):
        identifier = self.value
        symbol = symtable.lookup(identifier)
        if (not symbol):
            raise RuntimeException(f'{identifier} not defined')
        node = symbol.get_node()
        if (type(node) == FunctionDef):
            return StringType(node.get_type())
        return node.evaluate(symtable)


class FunctionCall(Expression):

    def __init__(self, name, expressions):
        self.name = name
        self.expressions = expressions
    
    def evaluate(self, symtable):
        identifier = self.name.get_identifier()
        symbol = symtable.lookup(identifier)
        if (not symbol):
            raise RuntimeException(f'{identifier} not defined')

        # Validate node type
        if (type(symbol.get_node()) != FunctionDef):
            raise RuntimeException(f'{identifier} is not a function')

        # Get function definition and its attributes
        functiondef = symbol.get_node()
        parameters = functiondef.get_parameters()
        body = functiondef.get_body()

        # Create child symbol table and insert parent symbols
        function_symtable = Function()
        symbols = symtable.to_dict()
        for _, s in symbols.items():
            function_symtable.insert(s.get_identifier(), s.get_node())

        # Validate parameters and expressions size
        if (len(parameters) != len(self.expressions)):
            raise RuntimeException(f'invalid parameters size for {identifier}')
        
        # Assign expression values to parameters respectively
        for parameter, expression in zip(parameters, self.expressions):
            arg_identifier = parameter.get_identifier()
            arg_node = Constant(expression.evaluate(symtable))
            function_symtable.insert(arg_identifier, arg_node)
        
        # Evaluate function body
        if (type(body) == list):
            # Case: user-defined function
            value = ASTUtil.evaluate_statements(
                function_symtable, body
            )
            if (type(value).__name__ == 'ReturnValue'):
                return value
            return ASTUtil.evaluate_base_type(None)
        else:
            # Case: built-in function
            function_pointer = body
            args = {}
            for parameter in parameters:
                value = parameter.evaluate(function_symtable)
                args[parameter.get_identifier()] = value
            return_value = function_pointer(**args)
            return ASTUtil.evaluate_base_type(return_value)


class Iterable(Constant): pass


class Tuple(Iterable):

    def __init__(self, expressions):
        self.expressions = expressions
    
    def evaluate(self, symtable):
        return ASTUtil.evaluate_base_type(tuple([
            expression.evaluate(symtable) 
            for expression in self.expressions
        ]))


class List(Iterable):

    def __init__(self, expressions):
        self.expressions = expressions
    
    def evaluate(self, symtable):
        return ASTUtil.evaluate_base_type([
            expression.evaluate(symtable) 
            for expression in self.expressions
        ])


class Dict(Iterable):

    def __init__(self, key_value_pairs):
        self.key_value_pairs = key_value_pairs

    def evaluate(self, symtable):
        the_dict = dict()
        for key_value_pair in self.key_value_pairs:
            key, value = key_value_pair.evaluate(symtable)
            the_dict[key] = value
        return ASTUtil.evaluate_base_type(the_dict)


class KeyValuePair(object):
    
    def __init__(self, expression_key, expression_value):
        self.expression_key = expression_key
        self.expression_value = expression_value

    def evaluate(self, symtable):
        key = self.expression_key.evaluate(symtable)
        value = self.expression_value.evaluate(symtable)
        return (key, value)


class IterableIndex(Expression):

    def __init__(self, name, expression):
        self.name = name
        self.expression = expression

    def evaluate(self, symtable):
        iterable = self.name.evaluate(symtable)
        index = self.expression.evaluate(symtable)
        return iterable[index]



class SyntaxException(Exception):

    def __init__(self, *args: object):
        super().__init__(*args)


class RuntimeException(Exception):
    
    def __init__(self, *args: object) -> None:
        super().__init__(*args)



import sys
from optparse import OptionParser
import src.sheng.compile as compile


def default(str):
    return str + ' [Default: %default]'

def read_command(argv):
    usage_str = 'sheng [option] [file]'
    parser = OptionParser(usage_str)

    parser.add_option('-d', '--debug', action='store_true', dest='debug',
                      help=default('debug mode'), default=False)
    
    options, other = parser.parse_args(argv[1:])
    if (len(other) == 0):
        parser.error('no input file')
    if (len(other) > 1):
        parser.error('Command line input not understood: ' + str(other[1:]))

    args = dict()
    args['file'] = other[0]
    args['debug'] = options.debug

    return args

def read_file(file, debug):
    f = open(file, "r", encoding='utf-8')
    data = f.read()
    f.close()
    return data, debug

def main():
    args = read_command(sys.argv)
    data, debug = read_file(**args)
    compile.exec(data, debug)
